/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef EVOLVE_HPP
#define EVOLVE_HPP

#include <algorithm>
#include <vector>
#include <numeric>
#include <cmath>
#include <cassert>
#include <functional>
#include <random>

namespace optimization {

/*
   The evolutionary optimization algorithm is implemented as a functor with
   assignable sub-functions.
   This allows most aspects of the algorithm to to be selected independently at
   runtime.
   The create(), procreate() and score() functions must be implemented. The
   other functions have
   default values, but the rank(), converged() and select() function will
   usually be customized. There
   is no restriction on the optimized type T, but the final optimized object is
   returned by value.
 */

template <class T>
struct evolutionary_optimizer {

  // functions with no default values
  std::function<void(T& member)> create;
  std::function<void(size_t member_index, std::vector<T>& population,
                     const std::vector<size_t>& indices, size_t nparents)>
      procreate;
  std::function<double(const T& member)> score;

  // functions with default values
  std::function<double(size_t score_index, const std::vector<double>& scores)>
      rank;
  std::function<void(int generation, const std::vector<T>& population,
                     const std::vector<double>& scores)> report;
  std::function<double(int generation)> select;
  std::function<bool(int generation, const std::vector<double>& scores)>
      converged;
  std::function<void(std::vector<T>& population,
                     const std::vector<size_t>& indices, size_t nsurv)>
      regenerate_population;
  std::function<void(
      const std::vector<T>& population, std::vector<double>& scores,
      const std::vector<size_t>& indices, size_t nsurv)> score_population;
  std::function<void(const std::vector<double>& scores,
                     std::vector<double>& ranks)> rank_population;

  static double rank_by_score(size_t index, const std::vector<double>& scores) {
    return scores[index];
  }
  static double rank_by_inverse_score(size_t index,
                                      const std::vector<double>& scores) {
    return -scores[index];
  }

  static double select_half(int generation) { return 0.5; }
  static void null_report(int, const std::vector<T>&,
                          const std::vector<double>&) {}

  evolutionary_optimizer();
  evolutionary_optimizer(const evolutionary_optimizer&) = default;
  evolutionary_optimizer& operator=(const evolutionary_optimizer&) = default;
  // Not yet supported on all platforms:
  // evolutionary_optimizer(evolutionary_optimizer&&) = default;
  // evolutionary_optimizer& operator=(evolutionary_optimizer&&) = default;
  virtual ~evolutionary_optimizer() = default;

  T operator()(size_t popsize);
};

template <class T>
evolutionary_optimizer<T>::evolutionary_optimizer() {
  using std::vector;
  score_population = [this](const vector<T>& population, vector<double>& scores,
                            const vector<size_t>& indices, size_t nsurv) {
    std::for_each(indices.cbegin() + nsurv, indices.cend(),
                  [&](size_t i) { scores[i] = score(population[i]); });
  };
  rank_population =
      [this](const vector<double>& scores, vector<double>& ranks) {
        for (auto i = size_t{0}; i < scores.size(); ++i) {
          ranks[i] = rank(i, scores);
        }
      };
  regenerate_population = [this](vector<T>& population,
                                 const vector<size_t>& indices, size_t nsurv) {
    if (nsurv < 1)
      std::for_each(population.begin(), population.end(), create);
    else
      std::for_each(indices.cbegin() + nsurv, indices.cend(), [&](size_t i) {
        procreate(i, population, indices, nsurv);
      });
  };
  rank = rank_by_score;
  report = null_report;
  select = select_half;
  converged = [](int generation, const vector<double>& scores) {
    return generation == 100;
  };
}

template <class T>
T evolutionary_optimizer<T>::operator()(size_t popsize) {
  using std::vector;
  auto pop = vector<T>(popsize);
  auto scores = vector<double>(popsize);
  auto ranks = vector<double>(popsize);
  auto indices = vector<size_t>(popsize);
  std::iota(indices.begin(), indices.end(), 0);
  auto gen = 0;
  auto nsurv = size_t{0};
  do {
    gen++;
    regenerate_population(pop, indices, nsurv);
    score_population(pop, scores, indices, nsurv);
    rank_population(scores, ranks);
    std::stable_sort(indices.begin(), indices.end(),
                     [&](size_t a, size_t b) { return ranks[a] < ranks[b]; });
    report(gen, pop, scores);
    nsurv = size_t(round(select(gen) * popsize));
    assert(nsurv >= 0 && nsurv <= popsize);
  } while (!converged(gen, scores));
  return pop[indices[0]];
}
}

#endif
