/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef FOURIERMAP_HPP
#define FOURIERMAP_HPP

#include "unitcell.hpp"
#include "fft.hpp"
#include "geometry.hpp"
#include "spacegroup.hpp"
#include <cmath>
#include <stdexcept>
#include <array>
#include <complex>

namespace crystallography {

template <typename RealT>
class fourier_map {

public:
  fourier_map(const unit_cell& cell, double res_low, double res_high,
              double sampling = 3.0);
  ~fourier_map();
  void add_atom(double x, double y, double z, double occ, double b, int atnum);
  void invert();

  std::complex<RealT> get_structure_factor(int h, int k, int l,
                                           double resolution) const;
  std::complex<RealT> get_structure_factor(int h, int k, int l) const {
    return get_structure_factor(h, k, l, cell.resolution(h, k, l));
  }
  const unit_cell& unitcell() const { return cell; }
  double low_res_limit() const { return res_low; }
  double high_res_limit() const { return res_high; }
  double is_inverted() const { return inverted; }

private:
  static int find_prime_factorable_value(int start_value, int maxprime,
                                         bool even_only, int increment,
                                         int max_increments);
  int calc_map_index(int x, int y, int z) const {
    return (nx * (ny * ((64 * nz + z) % nz) + (64 * ny + y) % ny) +
            (64 * nx + x) % nx);
  }

  const unit_cell& cell;
  double res_low;
  double res_high;
  int nx;
  int ny;
  int nz;
  RealT* real_part;
  RealT* imag_part;
  double samples;
  double smear;
  double scale;
  bool inverted;

  static const std::array<std::array<double, 9>, 99> form_factors;
};

template <typename RealT>
fourier_map<RealT>::fourier_map(const unit_cell& unitcell, double reslow,
                                double reshigh, double sampling)
    : cell(unitcell), res_low(reslow), res_high(reshigh), nx(0), ny(0), nz(0),
      real_part(0), imag_part(0), samples(sampling), smear(0.0), scale(0.0),
      inverted(false) {

  // determine the map dimensions from the cell dimensions and resolution
  constexpr auto MaxPrime = 7;
  constexpr auto IncrementSize = 2;
  constexpr auto MaxIncrements = 100;

  nx = (int)(0.5 + samples * cell.a_axis() / res_high);
  ny = (int)(0.5 + samples * cell.b_axis() / res_high);
  nz = (int)(0.5 + samples * cell.c_axis() / res_high);
  nx = find_prime_factorable_value(nx, MaxPrime, true, IncrementSize,
                                   MaxIncrements);
  ny = find_prime_factorable_value(ny, MaxPrime, true, IncrementSize,
                                   MaxIncrements);
  nz = find_prime_factorable_value(nz, MaxPrime, true, IncrementSize,
                                   MaxIncrements);

  const auto gridpoints = nx * ny * nz;
  real_part = new RealT[gridpoints]{};
  imag_part = new RealT[gridpoints]{};

  // use gaussian damping factor from Bricogne,
  // International Tables for Crystallography, Vol. B, p. 87
  // see also J. Navaza, Acta Cryst. (2002) A58, 568-573.
  constexpr auto lnq = 7.0;
  constexpr auto bmin = 10.0;
  auto oversample = samples * 0.5;
  smear = std::max((lnq * res_high * res_high /
                        std::max(oversample * (oversample - 1.0), 0.1) -
                    bmin),
                   0.0);
  scale = cell.volume() / (double)(gridpoints);
}

template <typename RealT>
fourier_map<RealT>::~fourier_map() {
  delete[] real_part;
  delete[] imag_part;
}

template <typename RealT>
void fourier_map<RealT>::add_atom(double x, double y, double z, double occ,
                                  double b, int atnum) {
  // if adding to an inverted map, reinitialize
  if (inverted) {
    for (auto i = 0; i < nx * ny * nz; ++i) {
      real_part[i] = 0;
      imag_part[i] = 0;
    }
    inverted = false;
  }
  if (occ < 0.00001)
    return;
  // calculate real-space form factors using reciprocal space coefficients
  const auto& ff = form_factors[atnum];
  const auto M_4PI = 4.0 * geom::PI;
  const auto M_4PI2 = 4.0 * geom::PI * geom::PI;
  auto ae1 = occ * ff[0] * pow(sqrt(M_4PI / (ff[1] + b + smear)), 3.0);
  auto ae2 = occ * ff[2] * pow(sqrt(M_4PI / (ff[3] + b + smear)), 3.0);
  auto ae3 = occ * ff[4] * pow(sqrt(M_4PI / (ff[5] + b + smear)), 3.0);
  auto ae4 = occ * ff[6] * pow(sqrt(M_4PI / (ff[7] + b + smear)), 3.0);
  auto ae5 = occ * ff[8] * pow(sqrt(M_4PI / (b + smear)), 3.0);
  auto be1 = M_4PI2 / (ff[4] + b + smear);
  auto be2 = M_4PI2 / (ff[5] + b + smear);
  auto be3 = M_4PI2 / (ff[6] + b + smear);
  auto be4 = M_4PI2 / (ff[7] + b + smear);
  auto be5 = M_4PI2 / (b + smear);
  // calculate a bounding box
  // first, the density at atom center
  auto rho0 = ae1 + ae2 + ae3 + ae4 + ae5;
  // find distance where electron density is below threshold
  auto dmax = 0.0;
  auto rho = 0.0;
  do {
    dmax += 0.1;
    auto dsq = dmax * dmax;
    rho = ae1 * exp(-be1 * dsq) + ae2 * exp(-be2 * dsq) +
          ae3 * exp(-be3 * dsq) + ae4 * exp(-be4 * dsq) + ae5 * exp(-be5 * dsq);
  } while (rho0 / rho < 200.0);

  // fractionalize the result
  auto fdmax = std::array<double, 3>{{dmax, dmax, dmax}};
  fdmax = cell.fractionalize(fdmax);
  // convert to a map distance
  auto mapmax = std::array<double, 3>{{fabs(fdmax[0]) * (double)nx,
                                       fabs(fdmax[1]) * (double)ny,
                                       fabs(fdmax[2]) * (double)nz}};

  // check against dmax^2 to avoid square root calculations
  auto dsqmax = dmax * dmax;

  const spacegroup& sg = cell.space_group();

  for (auto i = 0; i < sg.num_symm_ops(); ++i) {
    auto& symm = sg.symm_op(i);
    auto symm_pos = std::array<double, 3>{{x, y, z}};
    symm_pos = geom::transform(symm, symm_pos);
    auto atomcenter = cell.orthogonalize(symm_pos);

    symm_pos[0] *= (double)nx;
    symm_pos[1] *= (double)ny;
    symm_pos[2] *= (double)nz;

    auto low = std::array<int, 3>{{(int)(symm_pos[0] - mapmax[0] + 0.5),
                                   (int)(symm_pos[1] - mapmax[1] + 0.5),
                                   (int)(symm_pos[2] - mapmax[2] + 0.5)}};
    auto high = std::array<int, 3>{{(int)(symm_pos[0] + mapmax[0] + 0.5),
                                    (int)(symm_pos[1] + mapmax[1] + 0.5),
                                    (int)(symm_pos[2] + mapmax[2] + 0.5)}};

    // calculate electron density values at grid points
    auto fracpoint = std::array<double, 3>{};
    for (auto mx = low[0]; mx <= high[0]; mx++) {
      fracpoint[0] = (double)mx / (double)nx;
      for (auto my = low[1]; my <= high[1]; my++) {
        fracpoint[1] = (double)my / (double)ny;
        for (auto mz = low[2]; mz <= high[2]; mz++) {
          fracpoint[2] = (double)mz / (double)nz;
          auto point_in_cell = cell.orthogonalize(fracpoint);
          // calculate density if within cutoff distance
          auto dsq = geom::distsq(point_in_cell, atomcenter);
          if (dsq <= dsqmax) {
            rho = ae1 * exp(-be1 * dsq) + ae2 * exp(-be2 * dsq) +
                  ae3 * exp(-be3 * dsq) + ae4 * exp(-be4 * dsq) +
                  ae5 * exp(-be5 * dsq);
            real_part[calc_map_index(mx, my, mz)] += (float)rho;
          }
        }
      }
    }
  }
}

template <typename RealT>
void fourier_map<RealT>::invert() {
  auto d = std::array<int, 5>{};
  //  transform on x
  d[0] = nx * ny * nz;
  d[1] = 1;
  d[2] = d[0];
  d[3] = d[0];
  d[4] = nx;
  complex_fourier_transform(real_part, imag_part, nx, d.data());
  //  transform on y
  d[1] = nx;
  d[2] = nx * ny;
  d[3] = nx;
  d[4] = 1;
  complex_fourier_transform(real_part, imag_part, ny, d.data());
  //  transform on z
  d[1] = nx * ny;
  d[2] = d[0];
  d[3] = d[1];
  d[4] = 1;
  complex_fourier_transform(real_part, imag_part, nz, d.data());
  inverted = true;
}

template <typename RealT>
std::complex<RealT>
fourier_map<RealT>::get_structure_factor(int h, int k, int l,
                                         double resolution) const {
  if (resolution < res_high || resolution > res_low) {
    throw std::invalid_argument(
        "reflection out of resolution range in fourier_map");
  }
  auto i = calc_map_index(h, k, l);
  auto sf = std::complex<RealT>(real_part[i], -imag_part[i]);
  // corrections for smear and map volume
  auto sthol = 0.5 / resolution;
  return sf * RealT(scale * exp(smear * sthol * sthol));
}

// return smallest value >= test_value that is factorable by prime numbers <=
// max_prime
template <typename RealT>
int fourier_map<RealT>::find_prime_factorable_value(int test_value,
                                                    int max_prime,
                                                    bool even_only, int incr,
                                                    int max_increments) {
  if (even_only) {
    if (incr >= 0 && incr < 2)
      incr = 2;
    else if (incr == -1)
      incr = -2;
    if (incr >= 0)
      test_value = ((test_value + 1) / 2) * 2;
    else
      test_value = ((test_value - 1) / 2) * 2;
  }
  auto n = test_value;
  for (auto i = 0; i < max_increments; ++i) {
    for (auto j = 2; j <= max_prime; ++j) {
      while (n == (n / j) * j) {
        n = n / j;
      }
    }
    if (n == 1)
      return test_value;
    test_value += incr;
    n = test_value;
  }
  return 0;
}

// Cromer-Mann coefficients: a1, b1, a2, b2, a3, b3, a4, b4, c
template <typename RealT>
const std::array<std::array<double, 9>, 99> fourier_map<RealT>::form_factors = {
    {{{0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0}},
     {{0.4930020, 10.51090, 0.3229120, 26.12570, 0.1401910, 3.142360,
       4.0810000E-02, 57.79970, 3.0380001E-03}},
     {{0.8734000, 9.103700, 0.6309000, 3.356800, 0.3112000, 22.92760, 0.1780000,
       0.9821000, 6.3999998E-03}},
     {{1.128200, 3.954600, 0.7508000, 1.052400, 0.6175000, 85.39050, 0.4653000,
       168.2610, 3.7700001E-02}},
     {{1.591900, 43.64270, 1.127800, 1.862300, 0.5391000, 103.4830, 0.7029000,
       0.5420000, 3.8500000E-02}},
     {{2.054500, 23.21850, 1.332600, 1.021000, 1.097900, 60.34980, 0.7068000,
       0.1403000, -0.1932000}},
     {{2.310000, 20.84390, 1.020000, 10.20750, 1.588600, 0.5687000, 0.8650000,
       51.65120, 0.2156000}},
     {{12.21260, 5.7000001E-03, 3.132200, 9.893300, 2.012500, 28.99750,
       1.166300, 0.5826000, -11.52900}},
     {{3.048500, 13.27710, 2.286800, 5.701100, 1.546300, 0.3239000, 0.8670000,
       32.90890, 0.2508000}},
     {{3.539200, 10.28250, 2.641200, 4.294400, 1.517000, 0.2615000, 1.024300,
       26.14760, 0.2776000}},
     {{3.955300, 8.404200, 3.112500, 3.426200, 1.454600, 0.2306000, 1.125100,
       21.71840, 0.3515000}},
     {{4.762600, 3.285000, 3.173600, 8.842200, 1.267400, 0.3136000, 1.112800,
       129.4240, 0.6760000}},
     {{5.420400, 2.827500, 2.173500, 79.26110, 1.226900, 0.3808000, 2.307300,
       7.193700, 0.8584000}},
     {{6.420200, 3.038700, 1.900200, 0.7426000, 1.593600, 31.54720, 1.964600,
       85.08860, 1.115100}},
     {{6.291500, 2.438600, 3.035300, 32.33370, 1.989100, 0.6785000, 1.541000,
       81.69370, 1.140700}},
     {{6.434500, 1.906700, 4.179100, 27.15700, 1.780000, 0.5260000, 1.490800,
       68.16450, 1.114900}},
     {{6.905300, 1.467900, 5.203400, 22.21510, 1.437900, 0.2536000, 1.586300,
       56.17200, 0.8669000}},
     {{11.46040, 1.0400000E-02, 7.196400, 1.166200, 6.255600, 18.51940,
       1.645500, 47.77840, -9.557400}},
     {{7.484500, 0.9072000, 6.772300, 14.84070, 0.6539000, 43.89830, 1.644200,
       33.39290, 1.444500}},
     {{8.218600, 12.79490, 7.439800, 0.7748000, 1.051900, 213.1870, 0.8659000,
       41.68410, 1.422800}},
     {{8.626600, 10.44210, 7.387300, 0.6599000, 1.589900, 85.74840, 1.021100,
       178.4370, 1.375100}},
     {{9.189000, 9.021300, 7.367900, 0.5729000, 1.640900, 136.1080, 1.468000,
       51.35310, 1.332900}},
     {{9.759500, 7.850800, 7.355800, 0.5000000, 1.699100, 35.63380, 1.902100,
       116.1050, 1.280700}},
     {{10.29710, 6.865700, 7.351100, 0.4385000, 2.070300, 26.89380, 2.057100,
       102.4780, 1.219900}},
     {{10.64060, 6.103800, 7.353700, 0.3920000, 3.324000, 20.26260, 1.492200,
       98.73990, 1.183200}},
     {{11.28190, 5.340900, 7.357300, 0.3432000, 3.019300, 17.86740, 2.244100,
       83.75430, 1.089600}},
     {{11.76950, 4.761100, 7.357300, 0.3072000, 3.522200, 15.35350, 2.304500,
       76.88050, 1.036900}},
     {{12.28410, 4.279100, 7.340900, 0.2784000, 4.003400, 13.53590, 2.348800,
       71.16920, 1.011800}},
     {{12.83760, 3.878500, 7.292000, 0.2565000, 4.443800, 12.17630, 2.380000,
       66.34210, 1.034100}},
     {{13.33800, 3.582800, 7.167600, 0.2470000, 5.615800, 11.39660, 1.673500,
       64.81260, 1.191000}},
     {{14.07430, 3.265500, 7.031800, 0.2333000, 5.165200, 10.31630, 2.410000,
       58.70970, 1.304100}},
     {{15.23540, 3.066900, 6.700600, 0.2412000, 4.359100, 10.78050, 2.962300,
       61.41350, 1.718900}},
     {{16.08160, 2.850900, 6.374700, 0.2516000, 3.706800, 11.44680, 3.683000,
       54.76250, 2.131300}},
     {{16.67230, 2.634500, 6.070100, 0.2647000, 3.431300, 12.94790, 4.277900,
       47.79720, 2.531000}},
     {{17.00060, 2.409800, 5.819600, 0.2726000, 3.973100, 15.23720, 4.354300,
       43.81630, 2.840900}},
     {{17.17890, 2.172300, 5.235800, 16.57960, 5.637700, 0.2609000, 3.985100,
       41.43280, 2.955700}},
     {{17.35550, 1.938400, 6.728600, 16.56230, 5.549300, 0.2261000, 3.537500,
       39.39720, 2.825000}},
     {{17.17840, 1.788800, 9.643500, 17.31510, 5.139900, 0.2748000, 1.529200,
       164.9340, 3.487300}},
     {{17.56630, 1.556400, 9.818400, 14.09880, 5.422000, 0.1664000, 2.669400,
       132.3760, 2.506400}},
     {{17.77600, 1.402900, 10.29460, 12.80060, 5.726290, 0.1255990, 3.265880,
       104.3540, 1.912130}},
     {{17.87650, 1.276180, 10.94800, 11.91600, 5.417320, 0.1176220, 3.657210,
       87.66270, 2.069290}},
     {{17.61420, 1.188650, 12.01440, 11.76600, 4.041830, 0.2047850, 3.533460,
       69.79570, 3.755910}},
     {{3.702500, 0.2772000, 17.23560, 1.095800, 12.88760, 11.00400, 3.742900,
       61.65840, 4.387500}},
     {{19.13010, 0.8641320, 11.09480, 8.144870, 4.649010, 21.57070, 2.712630,
       86.84720, 5.404280}},
     {{19.26740, 0.8085200, 12.91820, 8.434670, 4.863370, 24.79970, 1.567560,
       94.29280, 5.378740}},
     {{19.29570, 0.7515360, 14.35010, 8.217580, 4.734250, 25.87490, 1.289180,
       98.60620, 5.328000}},
     {{19.33190, 0.6986550, 15.50170, 7.989290, 5.295370, 25.20520, 0.6058440,
       76.89860, 5.265930}},
     {{19.28080, 0.6446000, 16.68850, 7.472600, 4.804500, 24.66050, 1.046300,
       99.81560, 5.179000}},
     {{19.22140, 0.5946000, 17.64440, 6.908900, 4.461000, 24.70080, 1.602900,
       87.48250, 5.069400}},
     {{19.16240, 0.5476000, 18.55960, 6.377600, 4.294800, 25.84990, 2.039600,
       92.80290, 4.939100}},
     {{19.18890, 5.830300, 19.10050, 0.5031000, 4.458500, 26.89090, 2.466300,
       83.95710, 4.782100}},
     {{19.64180, 5.303400, 19.04550, 0.4607000, 5.037100, 27.90740, 2.682700,
       75.28250, 4.590900}},
     {{19.96440, 4.817420, 19.01380, 0.4208850, 6.144870, 28.52840, 2.523900,
       70.84030, 4.352000}},
     {{20.14720, 4.347000, 18.99490, 0.3814000, 7.513800, 27.76600, 2.273500,
       66.87760, 4.071200}},
     {{20.29330, 3.928200, 19.02980, 0.3440000, 8.976700, 26.46590, 1.990000,
       64.26580, 3.711800}},
     {{20.38920, 3.569000, 19.10620, 0.3107000, 10.66200, 24.38790, 1.495300,
       213.9040, 3.335200}},
     {{20.33610, 3.216000, 19.29700, 0.2756000, 10.88800, 20.20730, 2.695900,
       167.2020, 2.773100}},
     {{20.57800, 2.948170, 19.59900, 0.2444750, 11.37270, 18.77260, 3.287190,
       133.1240, 2.146780}},
     {{21.16710, 2.812190, 19.76950, 0.2268360, 11.85130, 17.60830, 3.330490,
       127.1130, 1.862640}},
     {{22.04400, 2.773930, 19.66970, 0.2220870, 12.38560, 16.76690, 2.824280,
       143.6440, 2.058300}},
     {{22.68450, 2.662480, 19.68470, 0.2106280, 12.77400, 15.88500, 2.851370,
       137.9030, 1.984860}},
     {{23.34050, 2.562700, 19.60950, 0.2020880, 13.12350, 15.10090, 2.875160,
       132.7210, 2.028760}},
     {{24.00420, 2.472740, 19.42580, 0.1964510, 13.43960, 14.39960, 2.896040,
       128.0070, 2.209630}},
     {{24.62740, 2.387900, 19.08860, 0.1942000, 13.76030, 13.75460, 2.922700,
       123.1740, 2.574500}},
     {{25.07090, 2.253410, 19.07980, 0.1819510, 13.85180, 12.93310, 3.545450,
       101.3980, 2.419600}},
     {{25.89760, 2.242560, 18.21850, 0.1961430, 14.31670, 12.66480, 2.953540,
       115.3620, 3.583240}},
     {{26.50700, 2.180200, 17.63830, 0.2021720, 14.55960, 12.18990, 2.965770,
       111.8740, 4.297280}},
     {{26.90490, 2.070510, 17.29400, 0.1979400, 14.55830, 11.44070, 3.638370,
       92.65660, 4.567960}},
     {{27.65630, 2.073560, 16.42850, 0.2235450, 14.97790, 11.36040, 2.982330,
       105.7030, 5.920460}},
     {{28.18190, 2.028590, 15.88510, 0.2388490, 15.15420, 10.99750, 2.987060,
       102.9610, 6.756210}},
     {{28.66410, 1.988900, 15.43450, 0.2571190, 15.30870, 10.66470, 2.989630,
       100.4170, 7.566720}},
     {{28.94760, 1.901820, 15.22080, 9.985190, 15.10000, 0.2610330, 3.716010,
       84.32980, 7.976280}},
     {{29.14400, 1.832620, 15.17260, 9.599900, 14.75860, 0.2751160, 4.300130,
       72.02900, 8.581540}},
     {{29.20240, 1.773330, 15.22930, 9.370460, 14.51350, 0.2959770, 4.764920,
       63.36440, 9.243540}},
     {{29.08180, 1.720290, 15.43000, 9.225900, 14.43270, 0.3217030, 5.119820,
       57.05600, 9.887500}},
     {{28.76210, 1.671910, 15.71890, 9.092270, 14.55640, 0.3505000, 5.441740,
       52.08610, 10.47200}},
     {{28.18940, 1.629030, 16.15500, 8.979480, 14.93050, 0.3826610, 5.675890,
       48.16470, 11.00050}},
     {{27.30490, 1.592790, 16.72960, 8.865530, 15.61150, 0.4179160, 5.833770,
       45.00110, 11.47220}},
     {{27.00590, 1.512930, 17.76390, 8.811740, 15.71310, 0.4245930, 5.783700,
       38.61030, 11.68830}},
     {{16.88190, 0.4611000, 18.59130, 8.621600, 25.55820, 1.482600, 5.860000,
       36.39560, 12.06580}},
     {{20.68090, 0.5450000, 19.04170, 8.448400, 21.65750, 1.572900, 5.967600,
       38.32460, 12.60890}},
     {{27.54460, 0.6551500, 19.15840, 8.707510, 15.53800, 1.963470, 5.525930,
       45.81490, 13.17460}},
     {{31.06170, 0.6902000, 13.06370, 2.357600, 18.44200, 8.618000, 5.969600,
       47.25790, 13.41180}},
     {{33.36890, 0.7040000, 12.95100, 2.923800, 16.58770, 8.793700, 6.469200,
       48.00930, 13.57820}},
     {{34.67260, 0.7009990, 15.47330, 3.550780, 13.11380, 9.556420, 7.025880,
       47.00450, 13.67700}},
     {{35.31630, 0.6858700, 19.02110, 3.974580, 9.498870, 11.38240, 7.425180,
       45.47150, 13.71080}},
     {{35.56310, 0.6631000, 21.28160, 4.069100, 8.003700, 14.04220, 7.443300,
       44.24730, 13.69050}},
     {{35.92990, 0.6464530, 23.05470, 4.176190, 12.14390, 23.10520, 2.112530,
       150.6450, 13.72470}},
     {{35.76300, 0.6163410, 22.90640, 3.871350, 12.47390, 19.98870, 3.210970,
       142.3250, 13.62110}},
     {{35.65970, 0.5890920, 23.10320, 3.651550, 12.59770, 18.59900, 4.086550,
       117.0200, 13.52660}},
     {{35.56450, 0.5633590, 23.42190, 3.462040, 12.74730, 17.83090, 4.807030,
       99.17220, 13.43140}},
     {{35.88470, 0.5477510, 23.29480, 3.415190, 14.18910, 16.92350, 4.172870,
       105.2510, 13.42870}},
     {{36.02280, 0.5293000, 23.41280, 3.325300, 14.94910, 16.09270, 4.188000,
       100.6130, 13.39660}},
     {{36.18740, 0.5119290, 23.59640, 3.253960, 15.64020, 15.36220, 4.185500,
       97.49080, 13.35730}},
     {{35.51030, 0.4986260, 22.57870, 2.966270, 12.77660, 11.94840, 4.921590,
       22.75020, 13.21160}},
     {{36.67060, 0.4836290, 24.09920, 3.206470, 17.34150, 14.31360, 3.493310,
       102.2730, 13.35920}},
     {{36.64880, 0.4651540, 24.40960, 3.089970, 17.39900, 13.43460, 4.216650,
       88.48340, 13.28870}},
     {{36.78810, 0.4510180, 24.77360, 3.046190, 17.89190, 12.89460, 4.232840,
       86.00300, 13.27540}},
     {{36.91850, 0.4375330, 25.19950, 3.007750, 18.33170, 12.40440, 4.243910,
       83.78810, 13.26740}}}};
}
#endif
